import { clone, pick, filterChange, hasLocalStorage, changesHandler as Changes, toPromise, uuid, } from 'pouchdb-utils';
import { isDeleted, isLocalId } from 'pouchdb-adapter-utils';
import { collectConflicts, traverseRevTree, latest as getLatest } from 'pouchdb-merge';
import { safeJsonParse, safeJsonStringify } from 'pouchdb-json';
import { MISSING_DOC, REV_CONFLICT, WSQ_ERROR, createError } from 'pouchdb-errors';
import { binaryStringToBlobOrBuffer as binStringToBlob, btoa } from 'pouchdb-binary-utils';
import parseHexString from './parseHex';
import websqlBulkDocs from './bulkDocs';
import { ADAPTER_VERSION, DOC_STORE, BY_SEQ_STORE, ATTACH_STORE, LOCAL_STORE, META_STORE, ATTACH_AND_SEQ_STORE, } from './constants';
import { qMarks, stringifyDoc, unstringifyDoc, select, compactRevs, websqlError, getSize, unescapeBlob, } from './utils';
import openDB from './openDatabase';
var websqlChanges = new Changes();
function fetchAttachmentsIfNecessary(document, opts, api, txn, cb) {
    var attachments = Object.keys(document._attachments || {});
    if (!attachments.length) {
        return cb && cb();
    }
    var numDone = 0;
    function checkDone() {
        if (++numDone === attachments.length && cb) {
            cb();
        }
    }
    function fetchAttachment(doc, att) {
        var attObj = doc._attachments[att];
        var attOpts = { binary: opts.binary, ctx: txn };
        api._getAttachment(doc._id, att, attObj, attOpts, function (_, data) {
            doc._attachments[att] = Object.assign(pick(attObj, ['digest', 'content_type']), { data: data });
            checkDone();
        });
    }
    attachments.forEach(function (att) {
        if (opts.attachments && opts.include_docs) {
            fetchAttachment(document, att);
        }
        else {
            document._attachments[att].stub = true;
            checkDone();
        }
    });
}
var POUCH_VERSION = 1;
var BY_SEQ_STORE_DELETED_INDEX_SQL = "CREATE INDEX IF NOT EXISTS 'by-seq-deleted-idx' ON " + BY_SEQ_STORE + ' (seq, deleted)';
var BY_SEQ_STORE_DOC_ID_REV_INDEX_SQL = "CREATE UNIQUE INDEX IF NOT EXISTS 'by-seq-doc-id-rev' ON " + BY_SEQ_STORE + ' (doc_id, rev)';
var DOC_STORE_WINNINGSEQ_INDEX_SQL = "CREATE INDEX IF NOT EXISTS 'doc-winningseq-idx' ON " + DOC_STORE + ' (winningseq)';
var ATTACH_AND_SEQ_STORE_SEQ_INDEX_SQL = "CREATE INDEX IF NOT EXISTS 'attach-seq-seq-idx' ON " + ATTACH_AND_SEQ_STORE + ' (seq)';
var ATTACH_AND_SEQ_STORE_ATTACH_INDEX_SQL = "CREATE UNIQUE INDEX IF NOT EXISTS 'attach-seq-digest-idx' ON " +
    ATTACH_AND_SEQ_STORE +
    ' (digest, seq)';
var DOC_STORE_AND_BY_SEQ_JOINER = BY_SEQ_STORE + '.seq = ' + DOC_STORE + '.winningseq';
var SELECT_DOCS = BY_SEQ_STORE +
    '.seq AS seq, ' +
    BY_SEQ_STORE +
    '.deleted AS deleted, ' +
    BY_SEQ_STORE +
    '.json AS data, ' +
    BY_SEQ_STORE +
    '.rev AS rev, ' +
    DOC_STORE +
    '.json AS metadata';
function WebSqlPouch(opts, cb) {
    var api = this;
    var size = getSize(opts);
    var idRequests = [];
    var instanceId = null;
    var encoding;
    api._name = opts.name;
    var websqlOpts = Object.assign({}, opts, {
        version: POUCH_VERSION,
        description: opts.name,
        size: size,
    });
    var openDBResult = openDB(websqlOpts);
    if (openDBResult.error) {
        return websqlError(cb)(openDBResult.error);
    }
    var db = openDBResult.db;
    if (typeof db.readTransaction !== 'function') {
        db.readTransaction = db.transaction;
    }
    api._db = db;
    function dbCreated() {
        if (hasLocalStorage()) {
            window.localStorage['_pouch__websqldb_' + api._name] = true;
        }
        cb(null, api);
    }
    function runMigration2(tx, callback) {
        tx.executeSql(DOC_STORE_WINNINGSEQ_INDEX_SQL);
        tx.executeSql('ALTER TABLE ' + BY_SEQ_STORE + ' ADD COLUMN deleted TINYINT(1) DEFAULT 0', [], function () {
            tx.executeSql(BY_SEQ_STORE_DELETED_INDEX_SQL);
            tx.executeSql('ALTER TABLE ' + DOC_STORE + ' ADD COLUMN local TINYINT(1) DEFAULT 0', [], function () {
                tx.executeSql("CREATE INDEX IF NOT EXISTS 'doc-store-local-idx' ON " + DOC_STORE + ' (local, id)');
                var sql = 'SELECT ' +
                    DOC_STORE +
                    '.winningseq AS seq, ' +
                    DOC_STORE +
                    '.json AS metadata FROM ' +
                    BY_SEQ_STORE +
                    ' JOIN ' +
                    DOC_STORE +
                    ' ON ' +
                    BY_SEQ_STORE +
                    '.seq = ' +
                    DOC_STORE +
                    '.winningseq';
                tx.executeSql(sql, [], function (tx, result) {
                    var deleted = [];
                    var local = [];
                    for (var i = 0; i < result.rows.length; i++) {
                        var item = result.rows.item(i);
                        var seq = item.seq;
                        var metadata = JSON.parse(item.metadata);
                        if (isDeleted(metadata)) {
                            deleted.push(seq);
                        }
                        if (isLocalId(metadata.id)) {
                            local.push(metadata.id);
                        }
                    }
                    tx.executeSql('UPDATE ' + DOC_STORE + 'SET local = 1 WHERE id IN ' + qMarks(local.length), local, function () {
                        tx.executeSql('UPDATE ' +
                            BY_SEQ_STORE +
                            ' SET deleted = 1 WHERE seq IN ' +
                            qMarks(deleted.length), deleted, callback);
                    });
                });
            });
        });
    }
    function runMigration3(tx, callback) {
        var local = 'CREATE TABLE IF NOT EXISTS ' + LOCAL_STORE + ' (id UNIQUE, rev, json)';
        tx.executeSql(local, [], function () {
            var sql = 'SELECT ' +
                DOC_STORE +
                '.id AS id, ' +
                BY_SEQ_STORE +
                '.json AS data ' +
                'FROM ' +
                BY_SEQ_STORE +
                ' JOIN ' +
                DOC_STORE +
                ' ON ' +
                BY_SEQ_STORE +
                '.seq = ' +
                DOC_STORE +
                '.winningseq WHERE local = 1';
            tx.executeSql(sql, [], function (tx, res) {
                var rows = [];
                for (var i = 0; i < res.rows.length; i++) {
                    rows.push(res.rows.item(i));
                }
                function doNext() {
                    if (!rows.length) {
                        return callback(tx);
                    }
                    var row = rows.shift();
                    var rev = JSON.parse(row.data)._rev;
                    tx.executeSql('INSERT INTO ' + LOCAL_STORE + ' (id, rev, json) VALUES (?,?,?)', [row.id, rev, row.data], function (tx) {
                        tx.executeSql('DELETE FROM ' + DOC_STORE + ' WHERE id=?', [row.id], function (tx) {
                            tx.executeSql('DELETE FROM ' + BY_SEQ_STORE + ' WHERE seq=?', [row.seq], function () {
                                doNext();
                            });
                        });
                    });
                }
                doNext();
            });
        });
    }
    function runMigration4(tx, callback) {
        function updateRows(rows) {
            function doNext() {
                if (!rows.length) {
                    return callback(tx);
                }
                var row = rows.shift();
                var doc_id_rev = parseHexString(row.hex, encoding);
                var idx = doc_id_rev.lastIndexOf('::');
                var doc_id = doc_id_rev.substring(0, idx);
                var rev = doc_id_rev.substring(idx + 2);
                var sql = 'UPDATE ' + BY_SEQ_STORE + ' SET doc_id=?, rev=? WHERE doc_id_rev=?';
                tx.executeSql(sql, [doc_id, rev, doc_id_rev], function () {
                    doNext();
                });
            }
            doNext();
        }
        var sql = 'ALTER TABLE ' + BY_SEQ_STORE + ' ADD COLUMN doc_id';
        tx.executeSql(sql, [], function (tx) {
            var sql = 'ALTER TABLE ' + BY_SEQ_STORE + ' ADD COLUMN rev';
            tx.executeSql(sql, [], function (tx) {
                tx.executeSql(BY_SEQ_STORE_DOC_ID_REV_INDEX_SQL, [], function (tx) {
                    var sql = 'SELECT hex(doc_id_rev) as hex FROM ' + BY_SEQ_STORE;
                    tx.executeSql(sql, [], function (tx, res) {
                        var rows = [];
                        for (var i = 0; i < res.rows.length; i++) {
                            rows.push(res.rows.item(i));
                        }
                        updateRows(rows);
                    });
                });
            });
        });
    }
    function runMigration5(tx, callback) {
        function migrateAttsAndSeqs(tx) {
            var sql = 'SELECT COUNT(*) AS cnt FROM ' + ATTACH_STORE;
            tx.executeSql(sql, [], function (tx, res) {
                var count = res.rows.item(0).cnt;
                if (!count) {
                    return callback(tx);
                }
                var offset = 0;
                var pageSize = 10;
                function nextPage() {
                    var sql = select(SELECT_DOCS + ', ' + DOC_STORE + '.id AS id', [DOC_STORE, BY_SEQ_STORE], DOC_STORE_AND_BY_SEQ_JOINER, null, DOC_STORE + '.id ');
                    sql += ' LIMIT ' + pageSize + ' OFFSET ' + offset;
                    offset += pageSize;
                    tx.executeSql(sql, [], function (tx, res) {
                        if (!res.rows.length) {
                            return callback(tx);
                        }
                        var digestSeqs = {};
                        function addDigestSeq(digest, seq) {
                            var seqs = (digestSeqs[digest] = digestSeqs[digest] || []);
                            if (seqs.indexOf(seq) === -1) {
                                seqs.push(seq);
                            }
                        }
                        for (var i = 0; i < res.rows.length; i++) {
                            var row = res.rows.item(i);
                            var doc = unstringifyDoc(row.data, row.id, row.rev);
                            var atts = Object.keys(doc._attachments || {});
                            for (var _i = 0, atts_1 = atts; _i < atts_1.length; _i++) {
                                var item = atts_1[_i];
                                var att = doc._attachments[item];
                                addDigestSeq(att.digest, row.seq);
                            }
                        }
                        var digestSeqPairs = [];
                        Object.keys(digestSeqs).forEach(function (digest) {
                            var seqs = digestSeqs[digest];
                            seqs.forEach(function (seq) {
                                digestSeqPairs.push([digest, seq]);
                            });
                        });
                        if (!digestSeqPairs.length) {
                            return nextPage();
                        }
                        var numDone = 0;
                        digestSeqPairs.forEach(function (pair) {
                            var sql = 'INSERT INTO ' + ATTACH_AND_SEQ_STORE + ' (digest, seq) VALUES (?,?)';
                            tx.executeSql(sql, pair, function () {
                                if (++numDone === digestSeqPairs.length) {
                                    nextPage();
                                }
                            });
                        });
                    });
                }
                nextPage();
            });
        }
        var attachAndRev = 'CREATE TABLE IF NOT EXISTS ' + ATTACH_AND_SEQ_STORE + ' (digest, seq INTEGER)';
        tx.executeSql(attachAndRev, [], function (tx) {
            tx.executeSql(ATTACH_AND_SEQ_STORE_ATTACH_INDEX_SQL, [], function (tx) {
                tx.executeSql(ATTACH_AND_SEQ_STORE_SEQ_INDEX_SQL, [], migrateAttsAndSeqs);
            });
        });
    }
    function runMigration6(tx, callback) {
        var sql = 'ALTER TABLE ' + ATTACH_STORE + ' ADD COLUMN escaped TINYINT(1) DEFAULT 0';
        tx.executeSql(sql, [], callback);
    }
    function runMigration7(tx, callback) {
        var sql = 'ALTER TABLE ' + DOC_STORE + ' ADD COLUMN max_seq INTEGER';
        tx.executeSql(sql, [], function (tx) {
            var sql = 'UPDATE ' +
                DOC_STORE +
                ' SET max_seq=(SELECT MAX(seq) FROM ' +
                BY_SEQ_STORE +
                ' WHERE doc_id=id)';
            tx.executeSql(sql, [], function (tx) {
                var sql = "CREATE UNIQUE INDEX IF NOT EXISTS 'doc-max-seq-idx' ON " + DOC_STORE + ' (max_seq)';
                tx.executeSql(sql, [], callback);
            });
        });
    }
    function checkEncoding(tx, callback) {
        tx.executeSql('SELECT HEX("a") AS hex', [], function (tx, res) {
            var hex = res.rows.item(0).hex;
            encoding = hex.length === 2 ? 'UTF-8' : 'UTF-16';
            callback();
        });
    }
    function onGetInstanceId() {
        while (idRequests.length > 0) {
            var idCallback = idRequests.pop();
            idCallback(null, instanceId);
        }
    }
    function onGetVersion(tx, dbVersion) {
        if (dbVersion === 0) {
            var meta_1 = 'CREATE TABLE IF NOT EXISTS ' + META_STORE + ' (dbid, db_version INTEGER)';
            var attach = 'CREATE TABLE IF NOT EXISTS ' +
                ATTACH_STORE +
                ' (digest UNIQUE, escaped TINYINT(1), body BLOB)';
            var attachAndRev = 'CREATE TABLE IF NOT EXISTS ' + ATTACH_AND_SEQ_STORE + ' (digest, seq INTEGER)';
            var doc = 'CREATE TABLE IF NOT EXISTS ' +
                DOC_STORE +
                ' (id unique, json, winningseq, max_seq INTEGER UNIQUE)';
            var seq_1 = 'CREATE TABLE IF NOT EXISTS ' +
                BY_SEQ_STORE +
                ' (seq INTEGER NOT NULL PRIMARY KEY AUTOINCREMENT, ' +
                'json, deleted TINYINT(1), doc_id, rev)';
            var local = 'CREATE TABLE IF NOT EXISTS ' + LOCAL_STORE + ' (id UNIQUE, rev, json)';
            tx.executeSql(attach);
            tx.executeSql(local);
            tx.executeSql(attachAndRev, [], function () {
                tx.executeSql(ATTACH_AND_SEQ_STORE_SEQ_INDEX_SQL);
                tx.executeSql(ATTACH_AND_SEQ_STORE_ATTACH_INDEX_SQL);
            });
            tx.executeSql(doc, [], function () {
                tx.executeSql(DOC_STORE_WINNINGSEQ_INDEX_SQL);
                tx.executeSql(seq_1, [], function () {
                    tx.executeSql(BY_SEQ_STORE_DELETED_INDEX_SQL);
                    tx.executeSql(BY_SEQ_STORE_DOC_ID_REV_INDEX_SQL);
                    tx.executeSql(meta_1, [], function () {
                        instanceId = uuid();
                        var initSeq = 'INSERT INTO ' + META_STORE + ' (db_version, dbid) VALUES (?,?)';
                        var initSeqArgs = [ADAPTER_VERSION, instanceId];
                        tx.executeSql(initSeq, initSeqArgs, function () {
                            onGetInstanceId();
                        });
                    });
                });
            });
        }
        else {
            var setupDone = function () {
                var migrated = dbVersion < ADAPTER_VERSION;
                if (migrated) {
                    tx.executeSql('UPDATE ' + META_STORE + ' SET db_version = ' + ADAPTER_VERSION);
                }
                var sql = 'SELECT dbid FROM ' + META_STORE;
                tx.executeSql(sql, [], function (tx, result) {
                    instanceId = result.rows.item(0).dbid;
                    onGetInstanceId();
                });
            };
            var tasks_1 = [
                runMigration2,
                runMigration3,
                runMigration4,
                runMigration5,
                runMigration6,
                runMigration7,
                setupDone,
            ];
            var i_1 = dbVersion;
            var nextMigration_1 = function (tx) {
                tasks_1[i_1 - 1](tx, nextMigration_1);
                i_1++;
            };
            nextMigration_1(tx);
        }
    }
    function setup() {
        db.transaction(function (tx) {
            checkEncoding(tx, function () {
                fetchVersion(tx);
            });
        }, websqlError(cb), dbCreated);
    }
    function fetchVersion(tx) {
        var sql = 'SELECT sql FROM sqlite_master WHERE tbl_name = ' + META_STORE;
        tx.executeSql(sql, [], function (tx, result) {
            if (!result.rows.length) {
                onGetVersion(tx, 0);
            }
            else if (!/db_version/.test(result.rows.item(0).sql)) {
                tx.executeSql('ALTER TABLE ' + META_STORE + ' ADD COLUMN db_version INTEGER', [], function () {
                    onGetVersion(tx, 1);
                });
            }
            else {
                tx.executeSql('SELECT db_version FROM ' + META_STORE, [], function (tx, result) {
                    var dbVersion = result.rows.item(0).db_version;
                    onGetVersion(tx, dbVersion);
                });
            }
        });
    }
    setup();
    function getMaxSeq(tx, callback) {
        var sql = 'SELECT MAX(seq) AS seq FROM ' + BY_SEQ_STORE;
        tx.executeSql(sql, [], function (tx, res) {
            var updateSeq = res.rows.item(0).seq || 0;
            callback(updateSeq);
        });
    }
    function countDocs(tx, callback) {
        var sql = select('COUNT(' + DOC_STORE + ".id) AS 'num'", [DOC_STORE, BY_SEQ_STORE], DOC_STORE_AND_BY_SEQ_JOINER, BY_SEQ_STORE + '.deleted=0');
        tx.executeSql(sql, [], function (tx, result) {
            callback(result.rows.item(0).num);
        });
    }
    api._remote = false;
    api.type = function () {
        return 'websql';
    };
    api._id = toPromise(function (callback) {
        callback(null, instanceId);
    });
    api._info = function (callback) {
        var seq;
        var docCount;
        db.readTransaction(function (tx) {
            getMaxSeq(tx, function (theSeq) {
                seq = theSeq;
            });
            countDocs(tx, function (theDocCount) {
                docCount = theDocCount;
            });
        }, websqlError(callback), function () {
            callback(null, {
                doc_count: docCount,
                update_seq: seq,
                websql_encoding: encoding,
            });
        });
    };
    api._bulkDocs = function (req, reqOpts, callback) {
        websqlBulkDocs(opts, req, reqOpts, api, db, websqlChanges, callback);
    };
    function latest(tx, id, rev, callback, finish) {
        var sql = select(SELECT_DOCS, [DOC_STORE, BY_SEQ_STORE], DOC_STORE_AND_BY_SEQ_JOINER, DOC_STORE + '.id=?');
        var sqlArgs = [id];
        tx.executeSql(sql, sqlArgs, function (a, results) {
            if (!results.rows.length) {
                var err = createError(MISSING_DOC, 'missing');
                return finish(err);
            }
            var item = results.rows.item(0);
            var metadata = safeJsonParse(item.metadata);
            callback(getLatest(rev, metadata));
        });
    }
    api._get = function (id, opts, callback) {
        var doc;
        var metadata;
        var tx = opts.ctx;
        if (!tx) {
            return db.readTransaction(function (txn) {
                api._get(id, Object.assign({ ctx: txn }, opts), callback);
            });
        }
        function finish(err) {
            callback(err, { doc: doc, metadata: metadata, ctx: tx });
        }
        var sql;
        var sqlArgs;
        if (!opts.rev) {
            sql = select(SELECT_DOCS, [DOC_STORE, BY_SEQ_STORE], DOC_STORE_AND_BY_SEQ_JOINER, DOC_STORE + '.id=?');
            sqlArgs = [id];
        }
        else if (opts.latest) {
            latest(tx, id, opts.rev, function (latestRev) {
                opts.latest = false;
                opts.rev = latestRev;
                api._get(id, opts, callback);
            }, finish);
            return;
        }
        else {
            sql = select(SELECT_DOCS, [DOC_STORE, BY_SEQ_STORE], DOC_STORE + '.id=' + BY_SEQ_STORE + '.doc_id', [BY_SEQ_STORE + '.doc_id=?', BY_SEQ_STORE + '.rev=?']);
            sqlArgs = [id, opts.rev];
        }
        tx.executeSql(sql, sqlArgs, function (a, results) {
            if (!results.rows.length) {
                var missingErr = createError(MISSING_DOC, 'missing');
                return finish(missingErr);
            }
            var item = results.rows.item(0);
            metadata = safeJsonParse(item.metadata);
            if (item.deleted && !opts.rev) {
                var deletedErr = createError(MISSING_DOC, 'deleted');
                return finish(deletedErr);
            }
            doc = unstringifyDoc(item.data, metadata.id, item.rev);
            finish();
        });
    };
    api._allDocs = function (opts, callback) {
        var results = [];
        var totalRows;
        var updateSeq;
        var start = 'startkey' in opts ? opts.startkey : false;
        var end = 'endkey' in opts ? opts.endkey : false;
        var key = 'key' in opts ? opts.key : false;
        var keys = 'keys' in opts ? opts.keys : false;
        var descending = 'descending' in opts ? opts.descending : false;
        var limit = 'limit' in opts ? opts.limit : -1;
        var offset = 'skip' in opts ? opts.skip : 0;
        var inclusiveEnd = opts.inclusive_end !== false;
        var sqlArgs = [];
        var criteria = [];
        var keyChunks = [];
        if (keys) {
            var distinctKeys_1 = [];
            keys.forEach(function (key) {
                if (distinctKeys_1.indexOf(key) === -1) {
                    distinctKeys_1.push(key);
                }
            });
            for (var index = 0; index < distinctKeys_1.length; index += 999) {
                var chunk = distinctKeys_1.slice(index, index + 999);
                if (chunk.length > 0) {
                    keyChunks.push(chunk);
                }
            }
        }
        else if (key !== false) {
            criteria.push(DOC_STORE + '.id = ?');
            sqlArgs.push(key);
        }
        else if (start !== false || end !== false) {
            if (start !== false) {
                criteria.push(DOC_STORE + '.id ' + (descending ? '<=' : '>=') + ' ?');
                sqlArgs.push(start);
            }
            if (end !== false) {
                var comparator = descending ? '>' : '<';
                if (inclusiveEnd) {
                    comparator += '=';
                }
                criteria.push(DOC_STORE + '.id ' + comparator + ' ?');
                sqlArgs.push(end);
            }
            if (key !== false) {
                criteria.push(DOC_STORE + '.id = ?');
                sqlArgs.push(key);
            }
        }
        if (!keys) {
            criteria.push(BY_SEQ_STORE + '.deleted = 0');
        }
        db.readTransaction(function (tx) {
            countDocs(tx, function (docCount) {
                totalRows = docCount;
            });
            if (opts.update_seq) {
                getMaxSeq(tx, function (theSeq) {
                    updateSeq = theSeq;
                });
            }
            if (limit === 0) {
                return;
            }
            if (keys) {
                var finishedCount_1 = 0;
                var allRows_1 = [];
                keyChunks.forEach(function (keyChunk) {
                    sqlArgs = [];
                    criteria = [];
                    var bindingStr = '';
                    keyChunk.forEach(function () {
                        bindingStr += '?,';
                    });
                    bindingStr = bindingStr.substring(0, bindingStr.length - 1);
                    criteria.push(DOC_STORE + '.id IN (' + bindingStr + ')');
                    sqlArgs = sqlArgs.concat(keyChunk);
                    var sql = select(SELECT_DOCS, [DOC_STORE, BY_SEQ_STORE], DOC_STORE_AND_BY_SEQ_JOINER, criteria, DOC_STORE + '.id ' + (descending ? 'DESC' : 'ASC'));
                    sql += ' LIMIT ' + limit + ' OFFSET ' + offset;
                    tx.executeSql(sql, sqlArgs, function (tx, result) {
                        finishedCount_1++;
                        for (var index = 0; index < result.rows.length; index++) {
                            allRows_1.push(result.rows.item(index));
                        }
                        if (finishedCount_1 === keyChunks.length) {
                            processResult(allRows_1);
                        }
                    });
                });
            }
            else {
                var sql = select(SELECT_DOCS, [DOC_STORE, BY_SEQ_STORE], DOC_STORE_AND_BY_SEQ_JOINER, criteria, DOC_STORE + '.id ' + (descending ? 'DESC' : 'ASC'));
                sql += ' LIMIT ' + limit + ' OFFSET ' + offset;
                tx.executeSql(sql, sqlArgs, function (tx, result) {
                    var rows = [];
                    for (var index = 0; index < result.rows.length; index++) {
                        rows.push(result.rows.item(index));
                    }
                    processResult(rows);
                });
            }
            function processResult(rows) {
                for (var i = 0, l = rows.length; i < l; i++) {
                    var item = rows[i];
                    var metadata = safeJsonParse(item.metadata);
                    var id = metadata.id;
                    var data = unstringifyDoc(item.data, id, item.rev);
                    var winningRev = data._rev;
                    var doc = {
                        id: id,
                        key: id,
                        value: { rev: winningRev },
                    };
                    if (opts.include_docs) {
                        doc.doc = data;
                        doc.doc._rev = winningRev;
                        if (opts.conflicts) {
                            var conflicts = collectConflicts(metadata);
                            if (conflicts.length) {
                                doc.doc._conflicts = conflicts;
                            }
                        }
                        fetchAttachmentsIfNecessary(doc.doc, opts, api, tx);
                    }
                    if (item.deleted) {
                        if (keys) {
                            doc.value.deleted = true;
                            doc.doc = null;
                        }
                        else {
                            continue;
                        }
                    }
                    if (!keys) {
                        results.push(doc);
                    }
                    else {
                        var index = keys.indexOf(id);
                        do {
                            results[index] = doc;
                            index = keys.indexOf(id, index + 1);
                        } while (index > -1 && index < keys.length);
                    }
                }
                if (keys) {
                    keys.forEach(function (key, index) {
                        if (!results[index]) {
                            results[index] = { key: key, error: 'not_found' };
                        }
                    });
                }
            }
        }, websqlError(callback), function () {
            var returnVal = {
                total_rows: totalRows,
                offset: opts.skip,
                rows: results,
            };
            if (opts.update_seq) {
                returnVal.update_seq = updateSeq;
            }
            callback(null, returnVal);
        });
    };
    api._changes = function (opts) {
        opts = clone(opts);
        if (opts.continuous) {
            var id_1 = api._name + ':' + uuid();
            websqlChanges.addListener(api._name, id_1, api, opts);
            websqlChanges.notify(api._name);
            return {
                cancel: function () {
                    websqlChanges.removeListener(api._name, id_1);
                },
            };
        }
        var descending = opts.descending;
        opts.since = opts.since && !descending ? opts.since : 0;
        var limit = 'limit' in opts ? opts.limit : -1;
        if (limit === 0) {
            limit = 1;
        }
        var results = [];
        var numResults = 0;
        function fetchChanges() {
            var selectStmt = DOC_STORE +
                '.json AS metadata, ' +
                DOC_STORE +
                '.max_seq AS maxSeq, ' +
                BY_SEQ_STORE +
                '.json AS winningDoc, ' +
                BY_SEQ_STORE +
                '.rev AS winningRev ';
            var from = DOC_STORE + ' JOIN ' + BY_SEQ_STORE;
            var joiner = DOC_STORE +
                '.id=' +
                BY_SEQ_STORE +
                '.doc_id' +
                ' AND ' +
                DOC_STORE +
                '.winningseq=' +
                BY_SEQ_STORE +
                '.seq';
            var criteria = ['maxSeq > ?'];
            var sqlArgs = [opts.since];
            if (opts.doc_ids) {
                criteria.push(DOC_STORE + '.id IN ' + qMarks(opts.doc_ids.length));
                sqlArgs = sqlArgs.concat(opts.doc_ids);
            }
            var orderBy = 'maxSeq ' + (descending ? 'DESC' : 'ASC');
            var sql = select(selectStmt, from, joiner, criteria, orderBy);
            var filter = filterChange(opts);
            if (!opts.view && !opts.filter) {
                sql += ' LIMIT ' + limit;
            }
            var lastSeq = opts.since || 0;
            db.readTransaction(function (tx) {
                tx.executeSql(sql, sqlArgs, function (tx, result) {
                    function reportChange(change) {
                        return function () {
                            opts.onChange(change);
                        };
                    }
                    for (var i = 0, l = result.rows.length; i < l; i++) {
                        var item = result.rows.item(i);
                        var metadata = safeJsonParse(item.metadata);
                        lastSeq = item.maxSeq;
                        var doc = unstringifyDoc(item.winningDoc, metadata.id, item.winningRev);
                        var change = opts.processChange(doc, metadata, opts);
                        change.seq = item.maxSeq;
                        var filtered = filter(change);
                        if (typeof filtered === 'object') {
                            return opts.complete(filtered);
                        }
                        if (filtered) {
                            numResults++;
                            if (opts.return_docs) {
                                results.push(change);
                            }
                            if (opts.attachments && opts.include_docs) {
                                fetchAttachmentsIfNecessary(doc, opts, api, tx, reportChange(change));
                            }
                            else {
                                reportChange(change)();
                            }
                        }
                        if (numResults === limit) {
                            break;
                        }
                    }
                });
            }, websqlError(opts.complete), function () {
                if (!opts.continuous) {
                    opts.complete(null, {
                        results: results,
                        last_seq: lastSeq,
                    });
                }
            });
        }
        fetchChanges();
    };
    api._close = function (callback) {
        if (db.close !== undefined) {
            console.log('closing sql database');
            db.close(function () { return callback(); }, function () { return callback(createError(WSQ_ERROR, 'failed_to_close')); });
        }
        else {
            callback();
        }
    };
    api._getAttachment = function (docId, attachId, attachment, opts, callback) {
        var res;
        var tx = opts.ctx;
        var digest = attachment.digest;
        var type = attachment.content_type;
        var sql = 'SELECT escaped, ' +
            'CASE WHEN escaped = 1 THEN body ELSE HEX(body) END AS body FROM ' +
            ATTACH_STORE +
            ' WHERE digest=?';
        tx.executeSql(sql, [digest], function (tx, result) {
            var item = result.rows.item(0);
            var data = item.escaped ? unescapeBlob(item.body) : parseHexString(item.body, encoding);
            if (opts.binary) {
                res = binStringToBlob(data, type);
            }
            else {
                res = btoa(data);
            }
            callback(null, res);
        });
    };
    api._getRevisionTree = function (docId, callback) {
        db.readTransaction(function (tx) {
            var sql = 'SELECT json AS metadata FROM ' + DOC_STORE + ' WHERE id = ?';
            tx.executeSql(sql, [docId], function (tx, result) {
                if (!result.rows.length) {
                    callback(createError(MISSING_DOC));
                }
                else {
                    var data = safeJsonParse(result.rows.item(0).metadata);
                    callback(null, data.rev_tree);
                }
            });
        });
    };
    api._doCompaction = function (docId, revs, callback) {
        if (!revs.length) {
            return callback();
        }
        db.transaction(function (tx) {
            var sql = 'SELECT json AS metadata FROM ' + DOC_STORE + ' WHERE id = ?';
            tx.executeSql(sql, [docId], function (tx, result) {
                var metadata = safeJsonParse(result.rows.item(0).metadata);
                traverseRevTree(metadata.rev_tree, function (isLeaf, pos, revHash, ctx, opts) {
                    var rev = pos + '-' + revHash;
                    if (revs.indexOf(rev) !== -1) {
                        opts.status = 'missing';
                    }
                });
                var sql = 'UPDATE ' + DOC_STORE + ' SET json = ? WHERE id = ?';
                tx.executeSql(sql, [safeJsonStringify(metadata), docId]);
            });
            compactRevs(revs, docId, tx);
        }, websqlError(callback), function () {
            callback();
        });
    };
    api._getLocal = function (id, callback) {
        db.readTransaction(function (tx) {
            var sql = 'SELECT json, rev FROM ' + LOCAL_STORE + ' WHERE id=?';
            tx.executeSql(sql, [id], function (tx, res) {
                if (res.rows.length) {
                    var item = res.rows.item(0);
                    var doc = unstringifyDoc(item.json, id, item.rev);
                    callback(null, doc);
                }
                else {
                    callback(createError(MISSING_DOC));
                }
            });
        });
    };
    api._putLocal = function (doc, opts, callback) {
        if (typeof opts === 'function') {
            callback = opts;
            opts = {};
        }
        delete doc._revisions;
        var oldRev = doc._rev;
        var id = doc._id;
        var newRev;
        if (!oldRev) {
            newRev = doc._rev = '0-1';
        }
        else {
            newRev = doc._rev = '0-' + (parseInt(oldRev.split('-')[1], 10) + 1);
        }
        var json = stringifyDoc(doc);
        var ret;
        function putLocal(tx) {
            var sql;
            var values;
            if (oldRev) {
                sql = 'UPDATE ' + LOCAL_STORE + ' SET rev=?, json=? ' + 'WHERE id=? AND rev=?';
                values = [newRev, json, id, oldRev];
            }
            else {
                sql = 'INSERT INTO ' + LOCAL_STORE + ' (id, rev, json) VALUES (?,?,?)';
                values = [id, newRev, json];
            }
            tx.executeSql(sql, values, function (tx, res) {
                if (res.rowsAffected) {
                    ret = { ok: true, id: id, rev: newRev };
                    if (opts.ctx) {
                        callback(null, ret);
                    }
                }
                else {
                    callback(createError(REV_CONFLICT));
                }
            }, function () {
                callback(createError(REV_CONFLICT));
                return false;
            });
        }
        if (opts.ctx) {
            putLocal(opts.ctx);
        }
        else {
            db.transaction(putLocal, websqlError(callback), function () {
                if (ret) {
                    callback(null, ret);
                }
            });
        }
    };
    api._removeLocal = function (doc, opts, callback) {
        if (typeof opts === 'function') {
            callback = opts;
            opts = {};
        }
        var ret;
        function removeLocal(tx) {
            var sql = 'DELETE FROM ' + LOCAL_STORE + ' WHERE id=? AND rev=?';
            var params = [doc._id, doc._rev];
            tx.executeSql(sql, params, function (tx, res) {
                if (!res.rowsAffected) {
                    return callback(createError(MISSING_DOC));
                }
                ret = { ok: true, id: doc._id, rev: '0-0' };
                if (opts.ctx) {
                    callback(null, ret);
                }
            });
        }
        if (opts.ctx) {
            removeLocal(opts.ctx);
        }
        else {
            db.transaction(removeLocal, websqlError(callback), function () {
                if (ret) {
                    callback(null, ret);
                }
            });
        }
    };
    api._destroy = function (opts, callback) {
        websqlChanges.removeAllListeners(api._name);
        db.transaction(function (tx) {
            var stores = [
                DOC_STORE,
                BY_SEQ_STORE,
                ATTACH_STORE,
                META_STORE,
                LOCAL_STORE,
                ATTACH_AND_SEQ_STORE,
            ];
            stores.forEach(function (store) {
                tx.executeSql('DROP TABLE IF EXISTS ' + store, []);
            });
        }, websqlError(callback), function () {
            if (hasLocalStorage()) {
                delete window.localStorage['_pouch__websqldb_' + api._name];
                delete window.localStorage[api._name];
            }
            callback(null, { ok: true });
        });
    };
}
export default WebSqlPouch;
//# sourceMappingURL=index.js.map