import { createError, WSQ_ERROR } from 'pouchdb-errors';
import { guardedConsole } from 'pouchdb-utils';
import { BY_SEQ_STORE, ATTACH_STORE, ATTACH_AND_SEQ_STORE } from './constants';
function escapeBlob(str) {
    return str
        .replace(/\u0002/g, '\u0002\u0002')
        .replace(/\u0001/g, '\u0001\u0002')
        .replace(/\u0000/g, '\u0001\u0001');
}
function unescapeBlob(str) {
    return str
        .replace(/\u0001\u0001/g, '\u0000')
        .replace(/\u0001\u0002/g, '\u0001')
        .replace(/\u0002\u0002/g, '\u0002');
}
function stringifyDoc(doc) {
    delete doc._id;
    delete doc._rev;
    return JSON.stringify(doc);
}
function unstringifyDoc(doc, id, rev) {
    var it = JSON.parse(doc);
    it._id = id;
    it._rev = rev;
    return it;
}
function qMarks(num) {
    var s = '(';
    while (num--) {
        s += '?';
        if (num) {
            s += ',';
        }
    }
    return s + ')';
}
function select(selector, table, joiner, where, orderBy) {
    return ('SELECT ' +
        selector +
        ' FROM ' +
        (typeof table === 'string' ? table : table.join(' JOIN ')) +
        (joiner ? ' ON ' + joiner : '') +
        (where ? ' WHERE ' + (typeof where === 'string' ? where : where.join(' AND ')) : '') +
        (orderBy ? ' ORDER BY ' + orderBy : ''));
}
function compactRevs(revs, docId, tx) {
    if (!revs.length) {
        return;
    }
    var numDone = 0;
    var seqs = [];
    function checkDone() {
        if (++numDone === revs.length) {
            deleteOrphans();
        }
    }
    function deleteOrphans() {
        if (!seqs.length) {
            return;
        }
        var sql = 'SELECT DISTINCT digest AS digest FROM ' +
            ATTACH_AND_SEQ_STORE +
            ' WHERE seq IN ' +
            qMarks(seqs.length);
        tx.executeSql(sql, seqs, function (tx, res) {
            var digestsToCheck = [];
            for (var i = 0; i < res.rows.length; i++) {
                digestsToCheck.push(res.rows.item(i).digest);
            }
            if (!digestsToCheck.length) {
                return;
            }
            var sql = 'DELETE FROM ' +
                ATTACH_AND_SEQ_STORE +
                ' WHERE seq IN (' +
                seqs
                    .map(function () {
                    return '?';
                })
                    .join(',') +
                ')';
            tx.executeSql(sql, seqs, function (tx) {
                var sql = 'SELECT digest FROM ' +
                    ATTACH_AND_SEQ_STORE +
                    ' WHERE digest IN (' +
                    digestsToCheck
                        .map(function () {
                        return '?';
                    })
                        .join(',') +
                    ')';
                tx.executeSql(sql, digestsToCheck, function (tx, res) {
                    var nonOrphanedDigests = new Set();
                    for (var i = 0; i < res.rows.length; i++) {
                        nonOrphanedDigests.add(res.rows.item(i).digest);
                    }
                    digestsToCheck.forEach(function (digest) {
                        if (nonOrphanedDigests.has(digest)) {
                            return;
                        }
                        tx.executeSql('DELETE FROM ' + ATTACH_AND_SEQ_STORE + ' WHERE digest=?', [digest]);
                        tx.executeSql('DELETE FROM ' + ATTACH_STORE + ' WHERE digest=?', [digest]);
                    });
                });
            });
        });
    }
    revs.forEach(function forEachRevision(rev) {
        var sql = 'SELECT seq FROM ' + BY_SEQ_STORE + ' WHERE doc_id=? AND rev=?';
        tx.executeSql(sql, [docId, rev], function (tx, res) {
            try {
                if (!res.rows.length) {
                    return checkDone();
                }
                var firstRow = res.rows.item(0);
                if (firstRow !== undefined) {
                    seqs.push(firstRow.seq);
                    tx.executeSql('DELETE FROM ' + BY_SEQ_STORE + ' WHERE seq=?', [firstRow.seq], checkDone);
                }
                else {
                    var args = {
                        type: 'ROW_IS_UNDEFINED',
                        sql: sql,
                        rows: res.rows,
                        rows_length: res.rows.length,
                    };
                    console.log('compactRevs::deleteOrphans[error: ROW_IS_UNDEFINED] -> ', args);
                }
            }
            catch (err) {
                var args = {
                    type: 'error',
                    sql: sql,
                    rows: res.rows,
                    rows_length: res.rows.length,
                };
                console.log('compactRevs::deleteOrphans[error: ' + err.message || 'unknown' + '] -> ', args);
                console.error(err);
            }
        });
    });
}
function websqlError(callback) {
    return function (event) {
        guardedConsole('error', 'WebSQL threw an error', event);
        var errorNameMatch = event && event.constructor.toString().match(/function ([^(]+)/);
        var errorName = (errorNameMatch && errorNameMatch[1]) || event.type;
        var errorReason = event.target || event.message;
        callback(createError(WSQ_ERROR, errorReason, errorName));
    };
}
function getSize(opts) {
    if ('size' in opts) {
        return opts.size * 1000000;
    }
    var isAndroid = typeof navigator !== 'undefined' && /Android/.test(navigator.userAgent);
    return isAndroid ? 5000000 : 1;
}
export { escapeBlob, unescapeBlob, stringifyDoc, unstringifyDoc, qMarks, select, compactRevs, getSize, websqlError, };
//# sourceMappingURL=utils.js.map